
/*
 * Programme de la station fixe pour P. Guivy
 * Auteur : P. Chour - 2018-02-09 - V1
 * Matériel : carte Arduino pro micro, afficheur I2C 2 lignes 16 caractères, Un interrupteur pour mode debug
 * Un bouton poussoir pour acquittement/reset Alarme
 */


/*
 * Ca vaut le coup de lire pourquoi il faut utiliser cette bibliothèque avec les Arduino... Pour plus d'info,
 * chercher avec "Arduino PinChangeInterrupt" sur le Web (voir en particulier https://github.com/NicoHood/PinChangeInterrupt)
 *  Arduino Uno/Nano/Mini: All pins are usable  
 *  Arduino Mega: 10, 11, 12, 13, 50, 51, 52, 53, A8 (62), A9 (63), A10 (64),
 *                A11 (65), A12 (66), A13 (67), A14 (68), A15 (69)
 *  Arduino Leonardo/Micro: 8, 9, 10, 11, 14 (MISO), 15 (SCK), 16 (MOSI)    <===== la carte utilisée ici
 *  HoodLoader2: All (broken out 1-7) pins are usable
 *  Attiny 24/44/84: All pins are usable
 *  Attiny 25/45/85: All pins are usable
 *  Attiny 13: All pins are usable
 *  Attiny 441/841: All pins are usable
 *  ATmega644P/ATmega1284P: All pins are usable
 */

#include <PinChangeInterrupt.h>
#include <PinChangeInterruptBoards.h>
#include <PinChangeInterruptPins.h>
#include <PinChangeInterruptSettings.h>
#include <EEPROM.h>


#include <Wire.h> 
#include <LiquidCrystal_I2C.h>
#include <TimerThree.h>
#include <pc_lex.h>

typedef struct TCtx TCtx;       // pour sauvegarde du contexte en EEPROM et variables en RAM

// Quelques constantes qui n'ont pas besoin d'occuper de la RAM ou de la flash...
#define bouton_acq  9           // A9 Pin pour le bouton ACQ
#define bouton_debug  19        // A1 Pin pour le bouton DEBUG
#define MaxTimer_Rx 10.0        // Nombre de secondes après lequel on considère qu'il y a un TimeOut en réception. Attention, si période > 10s, le time out est égal à la période !
#define bps_USB 9600            // vitesse de transmission pour les E/S asynchrone sur USB
#define bps_TxRx 76800          // vitesse de transmission pour les E/S asynchrones TxRx
#define PeriodeDefaut 1000000   // Période par défaut pour le timer3 (en micro secondes)

// Constantes sur l'afficheur
#define ColActivite 11          // Colonne sur l'afficheur où l'indicateur d'activité est affiché
#define ColAlarme 12            // Début où les alarmes R (réception), T (Température), V (Volt) et I (Intensité) sont affichées (4 caractères)

#define FacteurADCx1000 2.502443792766373  // 2560mV / 1023 = tension référence en mV / résolution ADC 

#define STX 2                   // STX = Start of Text en ASCII
#define ETX 3                   // ETX = End of Text en ASCII
#define ENQ 5                   // ENQ = demande de la part du fixe au mobile



// Quelques variables globales à l'application
//
// Mesures courantes
float MTemp = 0;
float MVolt = 0;
float MIntensite = 0;
// Dernière valeurs lues. Permet d'éviter des rafraichissements inutiles
float MOldTemp = 0; 
float MOldVolt = 0;
float MOldIntensite = 0;
// mémorisation des valeurs d'alarmes par type de mesure
float ATemp = 0;
float AVolt = 0;
float AIntensite = 0;
// Indication mesures valides (on en a décodé au moins une)
boolean BVTemp = false;
boolean BVVolt = false;
boolean BVIntensite = false;
// Booléen signalant une alarme pour chaque type de mesure
bool BATemp;
bool BAVolt;
bool BAIntensite;

// Timers
float     Timer_Rx = 0;         // Timer réception
float     Timer_ENQ = 0;        // Timer si mode interrogation pour réponse à ENQ
boolean   T1sec = false;        // booléen mis à vrai par le timer. Doit être remise à faux par l'utilisateur
boolean   BTimer_Rx = false;    // Indique une erreur en réception (Time out)
float     Periode = 1.0;        // Période d'horloge. Ne pas confondre avec Ctx.Periode

// Interface
boolean   BAffAlarme = false;   // vrai si l'afficheur affiche les valeurs en alarme, faux sinon (régime permanent)
boolean   BBouton_acq = false;  // vrai si on appuie sur le bouton Acq
boolean   BDebug = false;       // vrai si mode denug

const String SAlarm = "Alarme"; 
const char Tiret[7] = {'-','-','-','-',' ',' ','\0'};    
const char Depassement[7] = {'*','*','*','*',' ',' ','\0'};    
int       Activite = 0;         // Permet de faire évoluer un indicateur d'activité
const byte MagicValue[4] = {0x55,0xAA,0x66,0x99};

// Etat automate réception
#define Rx_STX 0
#define Rx_ETX 1
#define Rx_Trt 2
byte Rx_Etat = Rx_STX;          // Etat initial automate réception

byte SpcChar[8] = {0b00000,0b10000,0b01000,0b00100,0b00010,0b00001,0b00000}; // Définition d'un caractère spécial "\" pour l'indicateur d'activité

struct TCtx
{
  byte MagicValue[8];
  boolean BModeInterrogation;
  float Periode;
  // Limites hautes et basses par défaut pour chaque type de mesure. En deça ou au delà, il y a alarme  
  float TTempMin = 0;
  float TVoltMin = 0;
  float TIntensiteMin = 0;
  float TTempMax = 50;
  float TVoltMax = 5;
  float TIntensiteMax = 50;
  // Facteur de conversion
  float FTemp = 2.0;
  float FVolt = 2.0;
  float FIntensite = 20.0;  
};

TCtx Ctx;   // contexte des données de l'application qui doivent être stockées en EEPROM

/*
 * déclarations pour analyse lexicale
 */

LexBuf Buffer_USB;                // buffer pour réception et analyse lexicale sur USB
LexBuf Buffer_Rx;                 // buffer pour réception et analyse Lexicale sur Rx

LiquidCrystal_I2C lcd(0x27,16,2);  // Initialisation de l'afficheur : adresse 0x27, 16 caractères, 2 lignes



/*
 * Initialisation du programme
 */
void setup()
{
  int i;
  boolean EEPROMinit = true;
  byte *ptrCtx;
  lcd.init();                     // Initialisation de l'afficheur
  lcd.clear();
  lcd.backlight();                // Allumage rétro éclairage afficheur
  lcd.print("Initialisation");    // Un petit texte pour faire patienter
  lcd.setCursor(0,1);
  lcd.print("P. Chour V1.0");     // C'est moi, c'est moi, c'est moi !
  lcd.createChar(0, SpcChar);     // On créé dans l'afficheur le caractère dont on a besoin pour l'indicateur d'activité

  pinMode(bouton_acq, INPUT);     // La pin où est connecté le bouton ACQ en entrée. Cette pin est gérable par interruption
  pinMode(bouton_debug, INPUT);   // La pin où est connecté le bouton DEBUG. Cette pin n'est pas gérable par interruption
  digitalWrite(bouton_debug, HIGH); // On active la résistance de Pull Up.

  
  // Lecture (ou initialisation) du contexte en EEPROM
  for (i=0; i < sizeof(MagicValue); i++) {
    EEPROMinit = ((EEPROM.read(i) == MagicValue[i]) && EEPROMinit);
    Ctx.MagicValue[i] = MagicValue[i];
  }
  if (EEPROMinit) {   // Cas où l'EEPROM contient quelque chose de valide => On la lit
    ptrCtx = (void *)(&Ctx);
    for (i=0; i< sizeof(Ctx); i++) {
      (*ptrCtx) = EEPROM[i];
      ptrCtx += 1;    
    }
  } else {            // Cas où l'EEPROM ne contient pas quelque chose de valide => On l'initialise
    Ctx.BModeInterrogation = false; // on est en mode réception spontanné 
    Ctx.Periode = 1.0;          // Valeur par défaut = 1s
    Ctx.TTempMin = 0.0;         // temp min... = 0
    Ctx.TVoltMin = 0.0;         // Volt min = 0
    Ctx.TIntensiteMin = 0.0;    // Intensité min = 0
    Ctx.TTempMax = 50;          // Temp max = 50°
    Ctx.TVoltMax = 5;           // Volt max = 5V
    Ctx.TIntensiteMax = 50;     // IMax = 50mA
    Ctx.FTemp = 2.0;            // Coef conversion température = 2
    Ctx.FVolt = 2.0;            // Coef conversion tension = 2
    Ctx.FIntensite = 20.0;      // Coef conversion Intensité = 2
    UpdateCtx();
  }
  
  LexClear(&Buffer_USB);          // Initialisation analyseur lexical pour USB
  LexClear(&Buffer_Rx);           // Initialisation analyseur lexical pour Rx
  
  if (Ctx.BModeInterrogation) {   // Période d'horloge selon que l'on est en interrogation ou en attente réception spontannée
    Periode = Ctx.Periode;        // si mode interrogation, on prend la valeur sauvegardée dans le contexte
  } else {
    Periode = PeriodeDefaut / 1000000; // sinon, toujours 1 seconde
  }

  Timer3.initialize(Periode*1000000); // Horloge sous IT
  Timer3.attachInterrupt(TimerSec);   // La routine d'interruption est désormais attachée à l'interruption du timer 3
  
  Serial.begin(bps_USB);              // Initialisation ligne série sur USB (virtuelle)
  Serial1.begin(bps_TxRx,SERIAL_7N1); // Initialisation ligne série Tx et Rx (1 et 0 de la carte Pro Micro) : 7 bit, sans parité, 1 stop

  attachPinChangeInterrupt(digitalPinToPinChangeInterrupt(bouton_acq),InterAcq,RISING);  // Bouton ACQ sous interruption sur front montant. La routine de traitement est InterAcq.
  enablePinChangeInterrupt(digitalPinToPinChangeInterrupt(bouton_acq)); // On active l'interruption du bouton ACQ

  delay(1000);                  // Pour voir le texte initial sur l'afficheur
  ResetAlarme();                // Affichage initial : valeurs = ---- et pas d'alarme
}

/*
 * Procédure sous interruption
 * Positionne la variable d'état BBouton_acq à vrai pour signaler un appui sur le bouton ACQ
 * La variable doit être remise à faux par le programme
 */

void InterAcq()
{
  BBouton_acq = true;
}

/*
 * Routine d'interruption chaque seconde
*/
void TimerSec()
{
  T1sec = true;   // Signale que la routine s'est exécutée
}

/**********************************
 * ROUTINES APPLICATION
**********************************/
/*
 * Mise à jour du Contexte en EEPROM. Utilisation de la fonction EEPROM.Update qui n'écrit 
 * que si la donnée à mettre à jour est différente de celle qui est stockée.
 */
 void UpdateCtx()
 {
  int i;
  byte *ptrCtx;    
  ptrCtx = (void *)(&Ctx);
  for (i=0; i< sizeof(Ctx); i++) {
    EEPROM.update(i, byte(*(ptrCtx+i)));      
  }  
 }


/*
 * Met A dans B et B dans A si A > B
 */
void TriAB(float* A, float* B)
{
  float C;
  if (*A > *B) {
    C = *A;
    *A = *B;
    *B = C;
  }
}

/*
 * Traitementd de la commande Mode Interrogation (I, I OFF, I ON, I ON <valeur>)
 */

boolean DecodeInterrogation(LexBuf* B)
{
  String S1 = "";
  float V;
  boolean ok = true;
  LexFiltre(B);  
  while ((!(B->Lfin)) && (ok)) {
    S1 = LexIdent(B);
    if (S1=="ON") {
        LexFiltre(B);
        if (B->Lfin) {
          Ctx.Periode = (PeriodeDefaut / 1000000);
          Periode = Ctx.Periode;
          Timer3.setPeriod(PeriodeDefaut);
          Ctx.BModeInterrogation = true;
        } else {
          S1 = LexReal(B, &V);
          if (S1 == "") {
            ok = false;
          } else {
            if ((V>=0.1) && (V<=10.0)) {
              Ctx.Periode = V;
              Periode = Ctx.Periode;
              Timer3.setPeriod(long(V*1000000));
              Ctx.BModeInterrogation = true;
            } else {
              ok = false;
            }
          }
        }
    } else if (S1 == "OFF") {
      Timer3.setPeriod(PeriodeDefaut);
      Periode == (PeriodeDefaut / 1000000);      
      Ctx.BModeInterrogation = false;
    } else {
      ok = false;
    }
    LexFiltre(B);
  }
  UpdateCtx();
  return ok;
}
/*
 * Décodage du buffer de réception et stockage des valeurs dans les variables 
 * FTemp, FVolt, FIntensite
  */
boolean DecodeFacteur(LexBuf* B)
{
  String S1 = "";
  String S2 = "";  
  boolean ok  = true;
  float V;
  LexFiltre(B);
  while ((!(B->Lfin)) && (ok)) {
    S1 = LexIdent(B);
    if (S1 == ""){
      ok = false;
    } else {
      LexSkip(B,'=');
      S2 = LexReal(B, &V);
      if (S2 == "") {
        ok = false;
      } else {
        // Conversion des valeurs brutes des convertisseurs en valeur réelles selon 
        // les constantes de conversion. 
        switch (S1[0])
        {
          case 'T' : Ctx.FTemp = V;UpdateCtx();break;
          case 'V' : Ctx.FVolt = V;UpdateCtx();break;
          case 'I' : Ctx.FIntensite = V;UpdateCtx();break;
          default :  ok = false; break;
        }
      }
    }
    LexFiltre(B);  
  }
  UpdateCtx();
  return ok;
}


/* 
 *  décode le contenu du buffer pour une commande de type Alarm (A). Renvoie Talarm...Ttension si une valeur est trouvée
 *  dans la commande. Sinon, inchangé
 *  La structure de *B doit avoir été initialisée. 
 *  
 */
bool DecodeAlarm(LexBuf* B)
{
  bool ok = true;
  String S1 = "";
  String S2 = "";
  float V;
  LexFiltre(B);
  while ((!(B->Lfin)) && (ok)) {
      S1 = LexIdent(B);
      if (S1 != ""){
        LexSkip(B,'=');
        S2 = LexReal(B, &V);
        if (S2 == "") {
          ok = false; break;
        } else {
          if (S1 == "TMIN") {
            Ctx.TTempMin = V;
          } else if (S1 == "TMAX") {
            Ctx.TTempMax = V;
          } else if (S1 == "IMIN") {
            Ctx.TIntensiteMin = V;
          } else if (S1 == "IMAX") {
            Ctx.TIntensiteMax = V;
          } else if (S1 == "VMIN") {
            Ctx.TVoltMin = V;
          } else if (S1 == "VMAX") {
            Ctx.TVoltMax = V;
          } else {ok = false; break;}
        }
      } else {
        ok = false; break;
      }
      LexFiltre(B);  
  }
  TriAB(&Ctx.TTempMin,&Ctx.TTempMax);
  TriAB(&Ctx.TIntensiteMin,&Ctx.TIntensiteMax);
  TriAB(&Ctx.TVoltMin,&Ctx.TVoltMax);
  UpdateCtx();
  return ok;
}

/*
 * Décodage du buffer de réception et stockage des valeurs dans les variables 
 * TTemp, TVolt, TIntensite. Pas de séparateur attendus ni de '='
 * Met à jour l'indicateur de mesure valide
 * Renvoi faux si erreur dans le décodage
 * Pour mémoire, une trame vide n'est pas une erreur
 * A noter le caractère d'échappement '>' qui signale un commentaire.
  */
boolean DecodeRx(LexBuf* B)
{
  String S1 = "";
  String S2 = ""; 
  long   V;  
  LexStart(B);
  LexFiltre(B);
  while (!(B->Lfin)) {
    if (B->Lbuf[0] == '>') { // si code commentaire
      return true;           // on sort tout de suite 
    }
    S1 = LexIdent(B);
    if (S1 == ""){
      return false;
    }
    S2 = LexInt(B, &V);
    if (S2 == "") {
      return false;
    }
    // Conversion des valeurs brutes des convertisseurs en valeur réelles selon 
    // les constantes de conversion. 
    switch (S1[0])
    {
      case 'T' : MTemp = (Ctx.FTemp*FacteurADCx1000*(float)V)/1000; BVTemp = true; break;
      case 'V' : MVolt = (Ctx.FVolt*FacteurADCx1000*(float)V)/1000; BVVolt = true; break;
      case 'I' : MIntensite = (Ctx.FIntensite*FacteurADCx1000*(float)V)/1000; BVIntensite = true; break;
      default :  return false;
    }
    LexFiltre(B);  
  }
  return true;
}


/*
 * Converti un float en char et l'affiche en col x row sur l'afficheur
 * C = symbole à afficher en tête (T pour température, V pour Volt, I pour intensité...
 * A = valeur à afficher
 * ForceTirets à vrai : affichage de ---- à la place des valeurs
  * Si dépassement de capacité de l'affichage (6 digits max), affichage de ****
  */
void AffMesure(int col, int row, int precision, char C, float* A, boolean ForceTirets)
{
  int i;
  char B[20]; // buffer d'affichage (C,=,signe,7chiffres,signe,E,3 chiffres) 
  if (ForceTirets) {
    lcd.setCursor(col, row);
    lcd.print(C);
    lcd.print('=');
    lcd.print(Tiret);    
  } else {
    B[0]=C;     // Symbole dans buffer
    B[1]='=';   // suivi de '=' pour faire joli
    for(i=2;i<20;i++) {B[i]='\0';} // Buffer à Null
    dtostrf(*A,1,precision,B+2);  // conversion float en char
    lcd.setCursor(col, row);  // Positionnement curseur pour affichage
    if (B[8] == '\0') {
      i = 0;  // Affichage => Tout ce qui n'est pas Null.
      do
      {
        if (B[i] != '\0') {
          lcd.print(B[i]);
        } else {
          lcd.print(' ');
        }
        i = i+1;
      } while (i < 8);  
    } else {
      lcd.print(C);
      lcd.print('=');
      lcd.print(Depassement);
    }
  }
}

/*
 * Reset l'affichage des alarmes et les booléens correspondant. 
 * Réinitialise à faux les booléens qui définissent les valeurs valides
 */

void ResetAlarme()
{
  BAffAlarme = false;
  BATemp = false;
  BAVolt = false;
  BTimer_Rx = false;
  BAIntensite = false;
  lcd.clear();
  lcd.setCursor(ColAlarme, 1);      // Zone de visualisation des alarmes sur afficheur
  lcd.print(Tiret);                 // "----" = pas d'alarme en cours
  AffMesure(0,0,0,'T',&MTemp,(!BVTemp)); // RAZ de l'affichage. Valeurs de mesure non prises en compte
  AffMesure(8,0,0,'V',&MVolt,(!BVVolt));
  AffMesure(0,1,0,'I',&MIntensite,(!BVIntensite));  
  if (BDebug) {Serial.println("Reset "+SAlarm);}
}

/*
 * Affiche les valeurs en alarme
  * On efface l'écran, on affiche ALARM dans la zone d'alarme, on affiche les valeurs en
 * alarme pour celles qui le sont ou "---" pour celles qui ne le sont pas
 * Le booléen BAffAlarme permet d'éviter de réafficher à chaque appel. Il est réinitialisé
 * a Faux lors du Reset de l'alarme.
 */
void AffAlarme()
{
  if (!BAffAlarme) {
    BAffAlarme = true; // On est en affichage d'une alarme
    lcd.clear();
    lcd.setCursor(ColActivite, 1);
    lcd.print("ALARM");
    AffMesure(0,0,1,'T',&ATemp, (!BATemp));
    AffMesure(8,0,2,'V',&AVolt, (!BAVolt));
    AffMesure(0,1,0,'I',&AIntensite, (!BAIntensite));
  }
}

/*
 * Affiche les valeurs mesurées sur l'écran LCD et positionne les alarmes éventuelles
 * La procédure vérifie aussi si les conditions d'alarmes sont toujours valides : on a pu être
 * en alarme mais si l'utilisateur modifie les bornes en cours de route, l'alarme doit disparaitre
 * même si elle n'a pas été acquitée
 * En entrée :  ForceAff = true : Force l'affichage des valeurs T, V I courantes
 *              RAZAff = true : Réinitialise les valeurs de l'affichage (zone alarme à ---- et valeurs à ----
 */
void AfficheTVI(bool ForceAff)
{
  int i;
  // Si on est en affichage d'alarme, on attend un acquitement pour repasser en affichage normal
  if (BAffAlarme) {return;}
  if (BVTemp) {
    // Traitement des conditions d'alarme. Une alarme peut tomber si on change les conditions d'alarmes en cours de route
    if (BATemp) {
      if ((ATemp >= Ctx.TTempMin) && (ATemp <= Ctx.TTempMax)) {
        BATemp = false;
        lcd.setCursor(ColAlarme+1, 1); // On affiche le symbole correspondant à l'absence d'alarme dans la zone d'affichage des alarmes
        lcd.print('-');
      }
    }
    if ((MOldTemp != MTemp) || (ForceAff)) { // Si dernière mesure = mesure courant, on n'affiche pas sauf si forçage
      MOldTemp = MTemp;       // On mémorise la dernière valeur lue
      AffMesure(0,0,1,'T',&MTemp, false);
      if ((MTemp < Ctx.TTempMin) || (MTemp > Ctx.TTempMax)) {
        ATemp = MTemp;        // On mémorise la dernière valeur qui a généré une alarme
        BATemp = true;        // On mémorise que la mesure courante est en alarme
        if (BDebug) {     // Si mode debug, on envoie l'alarme à la console
          Serial.print(SAlarm+" T=");
          Serial.println(ATemp);
        }         
        lcd.setCursor(ColAlarme+1, 1); // On affiche le symbole correspondant à l'alarme dans la zone d'affichage des alarmes
        lcd.print('T');
      }
    }    
  }
  if (BVVolt) {
    if (BAVolt) {
      if ((AVolt >= Ctx.TVoltMin) && (AVolt <= Ctx.TVoltMax)) {
        BAVolt = false;
        lcd.setCursor(ColAlarme+2, 1); // On affiche le symbole correspondant à l'absence d'alarme dans la zone d'affichage des alarmes
        lcd.print('-');
      }
    }
    if ((MOldVolt != MVolt) || (ForceAff)) {
      MOldVolt = MVolt;
      AffMesure(8,0,2,'V',&MVolt, false);
      if ((MVolt < Ctx.TVoltMin) || (MVolt > Ctx.TVoltMax)) {
        AVolt = MVolt;
        BAVolt = true;
        if (BDebug) {     // Si mode debug, on envoie l'alarme à la console
          Serial.print(SAlarm+" V=");
          Serial.println(AVolt);
        }      
        lcd.setCursor(ColAlarme+2, 1);
        lcd.print('V');
      }
    }     
  }
  if (BVIntensite) {
    if (BAIntensite) {
      if ((AIntensite >= Ctx.TIntensiteMin) && (AIntensite <= Ctx.TIntensiteMax)) {
        BAIntensite = false;
        lcd.setCursor(ColAlarme+3, 1); // On affiche le symbole correspondant à l'absence d'alarme dans la zone d'affichage des alarmes
        lcd.print('-');
      }
    }
    if ((MOldIntensite != MIntensite) || (ForceAff)) {
      MOldIntensite = MIntensite;
      AffMesure(0,1,0,'I',&MIntensite, false);
      if ((MIntensite < Ctx.TIntensiteMin) || (MIntensite > Ctx.TIntensiteMax)) {
        AIntensite = MIntensite;
        BAIntensite = true;
        if (BDebug) {     // Si mode debug, on envoie l'alarme à la console
          Serial.print(SAlarm+" I=");
          Serial.println(AIntensite);
        }       
        lcd.setCursor(ColAlarme+3, 1);
        lcd.print('I');
      }
    }    
  }

  // Indicateur d'activité : successivement | / - chr(0) 
  lcd.setCursor(ColActivite, 1);
  Activite = (Activite+1)%4;
  switch (Activite)
  {
    case 0:lcd.print('|');break;
    case 1:lcd.print('/');break;
    case 2:lcd.print('-');break;
    case 3:lcd.write(0);break;
  }

  // Affichage alarme réception
  if (Timer_Rx >= MaxTimer_Rx) {
    if (!BTimer_Rx) {
      BTimer_Rx = true;
      lcd.setCursor(ColAlarme,1);
      lcd.print('R');
      if (BDebug) {Serial.println("Time Out Réception");}
    }
  } else {
    if (BTimer_Rx) {
      lcd.setCursor(ColAlarme,1);
      lcd.print('-');
      BTimer_Rx = false;
    }
  }  
}


/**********************************
 * APPLICATION
**********************************/

void loop()
{
  int i;
  int incomingByte = 0;
  String S = "";
 

    if (T1sec) {                // Ce qui doit être exécuté (que) chaque seconde
      if (Ctx.BModeInterrogation) { // si on est en mode interrogation
        if (Timer_ENQ == 0) {   // Si il n'y a aucun Timer en cours sur ENQ
          Serial1.print(char(ENQ)); // Alors, on émet un ENQ
        }
        Timer_ENQ += Periode;   // Timer_ENQ en cours 
        if (Timer_ENQ >= MaxTimer_Rx){ // s'il atteint la limite
          Timer_ENQ = 0;        // Alors, on le remet à 0. Au prochain coup, on réémetra un ENQ
          Timer_Rx = MaxTimer_Rx;   // On force la valeur du Timer_Rx à TimeOut pour signalement sur afficheur
        }
      } else {                  // On n'est pas en mode interrogation
        if (Timer_Rx < MaxTimer_Rx) {
          Timer_Rx += Periode;    // on incrémente le Timer_Rx pour signaler un éventuel Timeout
        }
      }
      AfficheTVI(false);

      // Traitement Bouton DEBUG
      // On teste le bouton chaque seconde. Du coup, bas besoin d'anti rebonds
      //
      if (digitalRead(bouton_debug) == HIGH) {  // Etat haut => mode debug
        if (!BDebug) {                          // si on est déjà en mode debug, on ne fait rien
          Serial.println("Debug On");           // Sinon, on affiche la prise en compte du mode debug
          BDebug = true;                        // et on passe en mode debug
        }
      } else {
        if (BDebug) {                           // Etat bas = fin mode debug. Même principe que passage en mode debug.
          //Serial.println("Debug Off");
          BDebug = false;
        }
      }
      T1sec = false;
    }

                                // Ce qui est exécuté par polling

    // Traitement Bouton ACQ
    // Si appuie sur bouton ACQ (traitement par interruption qui positionne BBouton_acq), on affiche les alarmes si elles ne sont
    // pas affichée ou on reset les alarmes si elles sont affichées
    if (BBouton_acq) {
      if (BAffAlarme) {
        ResetAlarme();
      } else {
        AffAlarme();
      }
      BBouton_acq = false;
    }

    // Traitement de la réception. Les états peuvent être : 
    // Attente STX : on filtre les caractères jusqu'à trouver un STX -> Passage dans l'état Attente ETX
    // Attente ETX : on stocke les caractères reçus dans un buffer. Si on rencontre un STX, on vide le buffer. Si on rencontre un ETX -> Passage dans l'état Traitement réception
    // Traitement réception : on décode et on traite les données reçues. Ce traitement terminé, on repasse dans l'état STX
    if (Serial1.available() > 0) {
      incomingByte = Serial1.read();
      switch (Rx_Etat)
      {
        case Rx_STX : 
          if (incomingByte == STX) {
            Rx_Etat = Rx_ETX;
            LexClear(&Buffer_Rx);
            break;
          }
        case Rx_ETX : 
          if (incomingByte == ETX) {
            Rx_Etat = Rx_Trt;
            break;
          }
          if (incomingByte == STX) {
            LexClear(&Buffer_Rx);
            break;
          }
          if (Buffer_Rx.Lbuf.length() < Lbufmax) {
            Buffer_Rx.Lbuf.concat((char)incomingByte);
            break;
          }
        case Rx_Trt:
          if (BDebug) {Serial.println(Buffer_Rx.Lbuf);}
          if (DecodeRx(&Buffer_Rx)) { // Si décodage ok
            Timer_ENQ = 0;    // Réinitialisation Timer_ENQ. 
            Timer_Rx = 0;     // Réinitialisation Timer_Rx
          }
          LexClear(&Buffer_Rx);
          Rx_Etat = Rx_STX;
          break;
      }
    }
    

    // Traitement Dialogue console (Serial, port série sur USB)
    if (Serial.available() > 0) {   // même si on est en mode debug, on vide l'UART USB au cas où...
          // read the incoming byte:
          incomingByte = Serial.read();
          if (BDebug) {             // On ne traite et on n'affiche qu'en mode debug
            if (Buffer_USB.Lbuf.length() < Lbufmax) {
              Buffer_USB.Lbuf.concat((char)incomingByte);
            }
            if (incomingByte == RC) {
              LexStart(&Buffer_USB);
              S = LexIdent(&Buffer_USB);
              if (S == "A") {       // Alarmes, si pas de paramètres, on affiche les valeurs, sinon, on lit les valeurs
                LexFiltre(&Buffer_USB);
                if (Buffer_USB.Lfin) {
                  Serial.print ("A ");
                  Serial.print(" T=[");
                  Serial.print(Ctx.TTempMin);
                  Serial.print(", ");
                  Serial.print(Ctx.TTempMax);
                  Serial.print("], V=[");
                  Serial.print(Ctx.TVoltMin);
                  Serial.print(", ");
                  Serial.print(Ctx.TVoltMax);
                  Serial.print("], I=[");
                  Serial.print(Ctx.TIntensiteMin);
                  Serial.print(", ");
                  Serial.print(Ctx.TIntensiteMax);
                  Serial.println("]");
                } else {
                  if (!DecodeAlarm(&Buffer_USB)) {
                    LexError(&Buffer_USB);
                  } 
                  AfficheTVI(true);        // si on a changé les bornes, il faut peut-être forcer l'affichage. Ca peut se produire même s'il y a erreur de décodage
                }
              } else if (S =="F") {
                LexFiltre(&Buffer_USB);
                if (Buffer_USB.Lfin) {
                  Serial.print ("F ");
                  Serial.print(" FT=");
                  Serial.print(Ctx.FTemp);
                  Serial.print(", FI=");
                  Serial.print(Ctx.FIntensite);
                  Serial.print(", FV=");
                  Serial.println(Ctx.FVolt);
                } else {
                  if (!DecodeFacteur(&Buffer_USB)) {
                    LexError(&Buffer_USB);
                  } 
                   AfficheTVI(true);         // si on a changé les facteurs, il faut peut-être forcer l'affichage.
                }
              } else if (S =="I") {
                LexFiltre(&Buffer_USB);
                if (Buffer_USB.Lfin) {
                  Serial.print ("I");
                  if (Ctx.BModeInterrogation) {
                    Serial.print(" ON");
                    Serial.print(" Période=");
                    Serial.println(Ctx.Periode);
                  } else {
                    Serial.println(" OFF");
                  }
                } else {
                  if (!DecodeInterrogation(&Buffer_USB)) {
                    LexError(&Buffer_USB);
                  }
                }
                
              } else if (S =="ACQ") {     // Acquitement alarme. Affiche les valeurs en alarme
                AffAlarme();
              } else if (S == "RESET") {  // acquitement et effacement des alarmes
                ResetAlarme();
              } else {
                LexError(&Buffer_USB);
              }
              LexClear(&Buffer_USB);
            }
          }
    }    


}

