/*
 * Programme de la station mobile pour P. Guivy
 * Auteur : P. Chour - 2018-02-09 - V1
 * Matériel : carte Arduino pro micro, un bouton pour reset
 */

/* 
 *  A propos des ADC sur le 32U4 : pour rappel, il y a au moins un bug dans la bibliothèque avr d'arduino et un dans la documentation 
 *  ATMEL.
 *  AnalogReference ne permet pas de positionner la tension de référence interne à 2560mV. Si on veut l'utiliser, il faut 
 *  programmer l'ADC manuellement.
 *  La doc ATMEL indique que MUX5 doit être positionné à 1 dans ADCSRB (c'est une note de bas de page) car MUX5 ne fait
 *  pas partie des bits d'ADMUS (le multiplexeur qui permet de sélectionner les ADC.
 *  Mais la doc indique que MUX5 est en lecture seule. C'est sans doute une typo
 *  Ce programme utilise les ADC8, 10 et 11 ce qui implique de positionner MUX5 à 1. Si on change d'ADC, il faut vérifier si 
 *  MUX5 doit être positionné ou pas.
 */

/*
 * Il y a deux modes de fonctionnement selon le define ci-dessous
 * define BasseConsommation vrai : la carte envoie une mesure par seconde et ne reçoit jamais.
 * define BasseConsommation faux : la carte peut recevoir des données : 
 *    ENQ : émet les mesures
 */

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/power.h>
#include <avr/wdt.h>
#include "prescaler.h"


// Caractères spéciaux pour émission/réception
#define ENQ 5                   // Demande mesure
#define STX 2                   // STX = Start of Text en ASCII
#define ETX 3                   // ETX = End of Text en ASCII

// Gestion horloges
//prescaler 32UA4 :
// 0x00 -> division par 1         0x04 -> division par 16       0x80 -> division par 256
// 0x01 -> division par 2         0x05 -> division par 32
// 0x02 -> division par 4         0x06 -> division par 64
// 0x03 -> division par 8         0x07 -> division par 128
// Définir les deux constantes ci-dessous 
//    Diviseur  = Diviseur d'horloge = 2^Prescaler
//    Prescaler = Prescaler correspondant au diviseur   Exemple : 0x02
// En pratique, il ne faut pas descendre en dessous de 8MHz (Prescaler = 0x01)car au delà l'erreur sur le débit de la voie
// série (à 76800 bps) devient trop importante (environ +/-8%) et la communication n'est plus possible. 
// Pour éviter ce problème, il faudrait négocier le débit (à bas débit) entre le mobile et le fixe puis
// en chosir un pour lesquel l'erreur est faible en fonction de la fréquence de fonctionnement du mobile.
// CE PROGRAMME NE GERE QUE LES FREQUENCES 16MHz, 8MHz et 4MHz (prescaler = 0, 1 et 2).


#define bps_TxRx 76800                  // vitesse de transmission pour les E/S asynchrones TxRx
#define Prescaler CLOCK_PRESCALER_2     // prescaler voir prescaler.h


// Interrupteur
#define Bouton_BasseConso  19        // A1 Pin pour le bouton DEBUG

// ADC
#define ADCTemp 0x00            // ADC8  sans Mux5
#define ADCVolt 0x02            // ADC10 sans Mux5
#define ADCIntensite 0x03       // ADC11 sans Mux5
#define PinADCTemp 4            // n° de Pin correspondant à l'ADC 8
#define PinADCVolt 6            // n° de Pin correspondant à l'ADC 10
#define PinADCIntensite 8       // n° de Pin correspondant à l'ADC 11

// Mesures courantes (lecture ADC)
int MTemp = 0;
int MVolt = 0;
int MIntensite = 0;
int OldMTemp = -1;              // -1 pas possible pour une mesure => Non initialisé
int OldMVolt = -1;              // Sert à optimiser la transmission : on ne transmet que les valeurs qui ont évolué.
int OldMIntensite = -1;

boolean BasseConsommation;
boolean ItWDT = false;          // Pour indiquer une IT du Watchdog. Mis a vrai dans la routine d'IT



/*
 * Initialisation du programme
 */
void setup()
{
  setClockPrescaler(Prescaler); // utilisation d'une fonction en assembleur inline pour initialiser le prescaler
  
  pinMode(Bouton_BasseConso, INPUT);     // La pin où est connecté le bouton DEBUG. Cette pin n'est pas gérable par interruption
  digitalWrite(Bouton_BasseConso, HIGH); // On active la résistance de Pull Up.
  BasseConsommation = (digitalRead(Bouton_BasseConso) == HIGH);
  pinMode(PinADCTemp, INPUT);
  pinMode(PinADCVolt, INPUT);
  pinMode(PinADCIntensite, INPUT);
  USART_Init_76800();        // Initialisation ligne série Tx et Rx (1 et 0 de la carte Pro Micro). 8 bits sans parité (défaut)
  power_timer3_disable();
  power_timer1_disable();
  power_spi_disable();  
  power_twi_disable();
  power_usb_disable(); 
  trueDelay(1000);
  InitWatchDog();
}


/**********************************
 * ROUTINES USART
**********************************/

/*
 * Initialisation USART 8 bits sans parité, asynchrone, vitesse 76800bps à la compilation. 
 */
void USART_Init_76800()
{
  unsigned int UBRR;
  UCSR1B = 0;                         // Désactive UART en émission/réception
  UCSR1D = 0;                         // pas de CTS ni de RTS
  UCSR1C = (1 << UCSZ11);             // 7 bits 1 stop pas de parité
  UCSR1A = (1 << U2X1);               // double vitesse. Permet d'avoir une meilleure précision en bps
  UBRR1H = 0;                         // poids fort du diviseur
  switch (Prescaler)
  {
    case CLOCK_PRESCALER_1 : UBRR1L = 25; break;
    case CLOCK_PRESCALER_2 : UBRR1L = 12; break;
    case CLOCK_PRESCALER_4 : UBRR1L = 6; break;
    default : UBRR1L = 6; break;      // non géré
  }
  UCSR1B = (1<<RXEN1)|(1<<TXEN1);     // active l'émission et la réception
}

/*
 * Réception par test d'état
 */
 unsigned char USART_Receive()
{
  while (!(UCSR1A & (1<<RXC1)));
  return UDR1;
}

/*
 * Indique si un caractère a été reçu
 */
boolean USART_Available()
{
  return (UCSR1A & (1<<RXC1));
}

/*
 * Attend jusqu'à ce que la transmission soit terminée (TXC1 = 1)
 */
boolean USART_Flush_Tx()
{
  while (!(UCSR1A & (1<<TXC1)));
  trueDelay(5); // On laisse le temps à la ligne de se stabiliser
}

/*
 * Emission d'un caractère
 */
void USART_Transmit_Char(char data)
{
  while (!( UCSR1A & (1<<UDRE1))); // Tant que le buffer n'est pas vide, on boucle
  UDR1 = data;  // transmission
}

/*
 * Emission d'un entier non signé
 */
void USART_Transmit_Int(int data)
{
  int i, j;
  byte v[4];
  if (!data) {
    USART_Transmit_Char('0');  // transmission
  } else {
    i = 0;
    while ((data) && (i < 4)) {
      v[i] = byte((data % 10) | 0x30);
      i += 1;
      data = data / 10;
    }
    for (j=(i-1); j >=0; j--) {
      USART_Transmit_Char((char)v[j]);
    }
  }
}


/*
 * Emission d'un Hexa
 */
void USART_Transmit_Hexa(int data)
{
  int i, j;
  byte v[4];
  if (!data) {
    USART_Transmit_Char('0');  // transmission
  } else {
    i = 0;
    while ((data) && (i < 4)) {
      v[i] = byte(data & 0x000F) | 0x30;
      if (v[i] > 0x39) {v[i] = v[i]+7;}
      i += 1;
      data = data >> 4;
    }
    for (j=(i-1); j >=0; j--) {
      USART_Transmit_Char((char)v[j]);
    }
  }
}

/*
 * Emission d'un code C (>Ex ou x est le code C)
 */
void USART_Transmit_Comment(char C)
{
  USART_Transmit_Char(char(STX));  
  USART_Transmit_Char('>');  
  USART_Transmit_Char('E');  
  USART_Transmit_Char(C);  
  USART_Transmit_Char(char(ETX)); 
}


/**********************************
 * ROUTINES SLEEP MMODE
**********************************/

/*
 * Routine d'interruption du Wtacdog
 */
ISR(WDT_vect)
{
  ItWDT = true;
  InitWatchDog();
}

/*
 * Activation du mode SLEEP
 */

void ActiveSleep(void)
{
  set_sleep_mode(SLEEP_MODE_IDLE);  // On autorise le mode IDLE
  noInterrupts();                                // Arrêt IT
  sleep_enable();                       // On autorise le mode SLEEP
  // Désactivation de certains périphériques
  power_adc_disable();                  // Désactive l'ADC
  power_usart1_disable(); 
  power_timer0_disable();
  setClockPrescaler(CLOCK_PRESCALER_128); // On passe à 16MHz / 128
  interrupts();                         // Réactive IT
  sleep_mode();                         // On déclenche le mode SLEEP
  
  noInterrupts();
  sleep_disable();                      // En sortie du mode SLEEP, on redémarre ici
  setClockPrescaler(Prescaler);  
  power_timer0_enable();
  power_usart1_enable(); 
  power_adc_enable();
  interrupts();
}

/*
 * Initialisation du wtachdog
 */
void InitWatchDog(void)
{
  noInterrupts();             // Désactive interruptions
  //wdt_reset();
  MCUSR &= ~(1<<WDRF);            // Réinitialise le flag IT Watchdog 
  WDTCSR |= (1<<WDCE) | (1<<WDE); // Pour écrire WDE (Watchdog Enable), il faut positionner WDCE (Watchdog Change Enable). On a 4 cycles pour initialiser WDTSCR.
  WDTCSR = (1<<WDIE) | (1<<WDE) | (0<<WDP3) | (1<<WDP2) | (1<<WDP1) | (0<<WDP0);   // 1 seconde, IT active
  interrupts();                   // Autorise Interruptions
}
/**********************************
 * ROUTINES APPLICATION
**********************************/

/*
 * Routine de lecture ADC Par test d'état
 */

void LectureADC() 
{
    int W;
  // Initialisation des convertisseurs
    ADCSRA = 0;   // Arrêt ADC
    switch (Prescaler) // division fréquence horloge par 128  => 125KHz (entre 50kHz et 200kHz d'après la doc).
    {
      case CLOCK_PRESCALER_1 : ADCSRA = (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);  break; 
      case CLOCK_PRESCALER_2 : ADCSRA = (1 << ADPS2) | (1 << ADPS1); break; 
      case CLOCK_PRESCALER_4 : UBRR1L = (1 << ADPS2) | (1 << ADPS0); break;
      default : ADCSRA = ADCSRA = (1 << ADPS2) | (1 << ADPS1); break;       // non géré, valeur au pif !
    }  
    ADCSRB |= (1 << MUX5);          // utilisation d'ADC8, 10 et 11
    ADMUX = (1<<REFS0) | (1<<REFS1) | ADCTemp;
    ADCSRA |= (1<<ADEN);            // Démarrage ADC
    trueDelay(2);
    ADCSRA |= 1<<ADSC;              // Démarrage conversion
    while (ADCSRA & (1 << ADSC));     // Attente que la conversion soit terminée    
    W = ADCL;
    MTemp = ((ADCH << 8) + W);
    ADMUX = (1<<REFS0) | (1<<REFS1) | ADCVolt;
    trueDelay(2);
    ADCSRA |= 1<<ADSC;              // Démarrage conversion
    while (ADCSRA & (1 << ADSC));     // Attente que la conversion soit terminée
    W = ADCL;
    MVolt = ((ADCH << 8) + W);
    ADMUX = (1<<REFS0) | (1<<REFS1) | ADCIntensite;
    trueDelay(2);
    ADCSRA |= 1<<ADSC;              // Démarrage conversion
    while (ADCSRA & (1 << ADSC));     // Attente que la conversion soit terminée
    W = ADCL;
    MIntensite = ((ADCH << 8) + W);
    ADCSRA &= ~(1<<ADEN);          // Arrêt du convertisseur pour ne pas trop consommer
}

/*
 * Lecture des mesures et émission. Procédure bloquante
 */
void LitEtEmetMesure()
{
  LectureADC();
  USART_Transmit_Char(char(STX));   // Même si les valeurs n'ont pas évolué, on transmet STX et ETX pour que la partie fixe sache que la partie mobile fonctionne
  if (MTemp != OldMTemp) {
    USART_Transmit_Char('T');
    USART_Transmit_Int(MTemp);
    OldMTemp = MTemp;
  }
  if (MVolt != OldMVolt) {
    USART_Transmit_Char('V');
    USART_Transmit_Int(MVolt);
    OldMVolt = MVolt;
  }
  if (MIntensite != OldMIntensite) {
    USART_Transmit_Char('I');
    USART_Transmit_Int(MIntensite);
    OldMIntensite = MIntensite;
  }
  USART_Transmit_Char(char(ETX));
}

/*
 * Application 
 */
void loop()
{
  if (!BasseConsommation) {       // Traitement lorsque l'on n'est pas en mode Basse Consommation
    if (digitalRead(Bouton_BasseConso) == HIGH) {  // Etat haut => On veut entrer en mode Basse Consommation
      BasseConsommation = true;
      trueDelay(200);                 // Anti rebond
    } else {                      // On est toujours pas en mode basse consommation. On attend les ENQ
      if (USART_Available()) {  // On attend pour ENQ (autres caractères filtrés
        if (USART_Receive() == ENQ) {  // Si ENQ, on transmet les mesures
          LitEtEmetMesure();
        }
      }
    }
  } else { // On est en mode basse consommation
    // mode basse consommation. Il faut tout initialiser à chaque réveil. 
    pinMode(Bouton_BasseConso, INPUT);     // La pin où est connecté le bouton Basse Consommation.
    digitalWrite(Bouton_BasseConso, HIGH); // On active la résistance de Pull Up.
    pinMode(PinADCTemp, INPUT);
    pinMode(PinADCVolt, INPUT);
    pinMode(PinADCIntensite, INPUT);
    USART_Init_76800();        // Initialisation ligne série Tx et Rx (1 et 0 de la carte Pro Micro)
    if (digitalRead(Bouton_BasseConso) == LOW) {  // Etat bas => On veut sortir du mode basse consommation
      BasseConsommation = false;
      trueDelay(200);                 // Anti rebond
    } else {
      if (ItWDT) {
        LitEtEmetMesure();
        USART_Flush_Tx();
        ItWDT = false;
      }
      ActiveSleep();
    }
  }
}

